/*
 * ProtoDemo
 *
 *  Created on: 21.03.2014
 *      Author: vladimir
 */

#ifdef WIN32
#include <windows.h>
#include <winbase.h>
#else
#include <termios.h>
#endif

#include <stdio.h>
#include <fcntl.h>
#include <string>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#ifdef WIN32
typedef HANDLE Port;
#else
typedef int Port;
#endif


enum {
	PROTO_CMD_PING = 1,
	PROTO_CMD_DISPENSE = 2,
	PROTO_CMD_STATUS = 3,
};

const unsigned char proto_sync = 0x55;
const unsigned char proto_base_size = 7;

typedef struct {
	unsigned char sync;
	unsigned char size;
	unsigned char tmp1;
	unsigned char tmp2;
	unsigned char cmd;
	unsigned char cmd_data[48 + 2];
} Packet;


static unsigned int GetCRC16 (const unsigned char* bufData, unsigned int sizeData) {
	const int polynomial = 0x08408;
	unsigned int CRC, i;
	unsigned char j;
	CRC = 0;
	for (i = 0; i < sizeData; i++) {
		CRC ^= bufData[i];
		for (j = 0; j < 8; j++) {
			if (CRC & 0x0001) {
				CRC >>= 1;
				CRC ^= polynomial;
			}
			else CRC >>= 1;
		}
	}
	return CRC;
}

unsigned short CalcCRC (Packet* packet) {
	return GetCRC16 ((const unsigned char*) packet, packet->size - 2);
}

void SetCRC (Packet* packet) {
	*((unsigned short*) ((unsigned char*) packet + packet->size - 2)) =
		CalcCRC (packet);
}

unsigned short GetCRC (Packet* packet) {
	return *((unsigned short*) ((unsigned char*) packet + packet->size - 2));
}


std::string DumpHex (Packet* packet) {
	std::string str;
	unsigned int len = packet->size;
	unsigned char* data = (unsigned char*) packet;
	unsigned int i = 0;
	while (true) {
		if ((i > 0 && !(i % 16)) || i >= len) {
			if (i >= len) {
				for (int j = i; j % 16; j++)
					str += "   ";
			}
			str += "| ";
			static const unsigned char read_chars[] =
				{'+', '-', '=', '_', '/', '\\', '\'', '"', '.', ',', ' '};
			for (unsigned int j = i - ((i - 1) % 16) - 1; j < i; j++) {
				bool is_read_char =
						('a' <= data[j] && data[j] <= 'z') ||
						('A' <= data[j] && data[j] <= 'Z') ||
						('0' <= data[j] && data[j] <= '9');

				for (int k = 0; k < sizeof (read_chars); k++) {
					if (data[j] == read_chars[k]) {
						is_read_char = true;
						break;
					}
				}
				if (is_read_char)
					str += data[j];
				else
					str += ".";
			}

			if (i >= len)
				break;
			str += "\n";
		}
		char buf[8];
		sprintf (buf, "%.2X ", data[i]);
		str += buf;
		i++;
	}
	return str;
}

Packet* ParsePacket (unsigned char** p, unsigned int* size) {
	Packet* packet;
	while (*size > 0) {
		if (**p != proto_sync) {
			(*p)++; (*size)--;
			continue;
		}
		if (*size < 5)
			return NULL;

		packet = (Packet*) *p;
		if (packet->size > 64) {
			(*p)++; (*size)--;
			continue;
		}
		if (packet->size > *size)
			return NULL;

		if (GetCRC (packet) != CalcCRC (packet)) {
			(*p)++; (*size)--;
			continue;
		}
		(*p) += packet->size;
		(*size) -= packet->size;
		return packet;
	}
	return NULL;
}


bool SendPacket (Port port, Packet* packet) {
	packet->sync = proto_sync;
	packet->tmp1 = 0;
	packet->tmp2 = 0;

	SetCRC (packet);

	printf ("-> %s\n", DumpHex (packet).c_str ());
#ifdef WIN32
	DWORD bytes = 0;
	if (!WriteFile (port, (char*) packet, packet->size, &bytes, NULL) || bytes != packet->size) {
		printf ("Failed to write to port %d\n", bytes);
		return false;
	}

#else
	if (write (port, (char*) packet, packet->size) != packet->size) {
		printf ("Failed to write to port\n");
		return false;
	}

#endif
	return true;
}

bool RecvData (Port port, char* data, int size) {

	while (size) {

#ifdef WIN32
		DWORD bytes;

		if (!ReadFile (port, (char*) data, size, &bytes, NULL) || bytes == 0) {
			printf ("Failed to read from port\n");
			return false;
		}
#else
		int bytes;
		bytes = read (port, (char*) data, size);
		if (bytes <= 0) {
			printf ("Failed to read from port\n");
			return false;
		}
#endif
		size -= bytes;
		data += bytes;
	}
	return true;
}

bool RecvPacket (Port port, Packet* packet) {

	// Считываем заголовок пакета
	if (!RecvData (port, (char*) packet, 4) || packet->sync != proto_sync || packet->size > 64) {
		printf ("Out of sync\n");
		return false;
	}
	// Считываем оставшуюся часть
	if (!RecvData (port, ((char*) packet) + 4, packet->size - 4)) {
		printf ("No data in packet\n");
		return false;
	}

	if (GetCRC (packet) != CalcCRC (packet)) {
		printf ("Bad CRC: %X  != %X\n", GetCRC (packet), CalcCRC (packet));
		return false;
	}

	if (packet->cmd == 0) {
		printf ("Error in response: %X\n", packet->cmd_data[0]);
		return false;
	}
	printf ("<- %s\n", DumpHex (packet).c_str ());

	return true;
}


void usage () {
	printf (
		"Usage: protodemo <port>											\n"
		"								ping								\n"
		"								dispense <count>                    \n"
		"								status								\n\n"
		"Automated: http://www.stampsoft.ru\n"
	);
}

int main (int argc, char* argv[]) {
	if (argc < 3) {
		usage ();
		return 1;
	}

	// Открытие порта
	Port port;
#ifdef WIN32
	std::string path = "\\\\.\\";
	path += argv[1];
	port = CreateFile (path.c_str (), GENERIC_READ | GENERIC_WRITE,
		0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (port == INVALID_HANDLE_VALUE) {
		printf ("Failed to open port");
		return 1;
	}


	DCB dcb;
	if (!GetCommState (port, &dcb)) {
		printf ("Failed to get COM state");
		CloseHandle (port);
		return 2;
	}

	dcb.BaudRate = CBR_19200;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;

	dcb.ByteSize = 8;

	if (!SetCommState (port, &dcb)) {
		printf ("Failed to set COM state");
		CloseHandle (port);
		return 4;
	}

	PurgeComm (port, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR);
	SetupComm (port,256,256);


#else
	port = open (argv[1], O_RDWR | O_NOCTTY);
	if (port < 0) {
		perror ("Failed to open port");
		return 1;
	}
	struct termios opt;

	if (tcgetattr (port, &opt) < 0) {
		return 3;
	}

	opt.c_oflag = 0;
	opt.c_lflag = 0;
	opt.c_iflag = IGNPAR;
	opt.c_cflag |= CREAD | CSTOPB | CLOCAL | CS8;
	opt.c_cflag &= ~CBAUD;
	opt.c_cflag |= B19200;


	if (tcsetattr (port, TCSANOW, &opt) < 0) {
		return 2;
	}

	tcflush (port, TCIFLUSH);

#endif

	// Идентификация устройства и его характеристик
	Packet packet;

	std::string cmd = argv[2];

	if (cmd == "ping") {
		Packet packet;
		packet.cmd = PROTO_CMD_PING;
		packet.size = proto_base_size;
		if (!SendPacket (port, &packet) || !RecvPacket (port, &packet)) {
			printf ("Failed to ping device\n");
			return 1;
		}
		printf ("PING %s\n", (packet.cmd == PROTO_CMD_PING ? "OK" : "FAILED"));

	} else if (cmd == "dispense" && argc == 4) {
		int count = atoi (argv[3]);

		Packet packet;
		packet.cmd = PROTO_CMD_DISPENSE;
		packet.size = proto_base_size + 1;
		packet.cmd_data[0] = count;

		if (!SendPacket (port, &packet) || !RecvPacket (port, &packet)) {
			printf ("Failed to move to opened position\n");
			return 1;
		}

	} else if (cmd == "status") {
		Packet packet;
		packet.cmd = PROTO_CMD_STATUS;
		packet.size = proto_base_size;

		if (!SendPacket (port, &packet) || !RecvPacket (port, &packet)) {
			printf ("Failed to fetch status\n");
			return 1;
		}
		unsigned char state = packet.cmd_data[0];
		const char* state_str = "UNKNOWN";
		if (state == 0)
			state_str = "WAIT INIT";
		else if (state == 1)
			state_str = "IDLE";
		else if (state == 2)
			state_str = "IDLE";
		else if (state == 3)
			state_str = "JAMMED";

		unsigned char dispense_pending = packet.cmd_data[1];
		unsigned char low = packet.cmd_data[2];

		printf ("Status: %d [%s], dispense_pending: %d, loaded: [%s]\n", state, state_str, dispense_pending, (low ? "LOW" : "OK"));

	} else
		usage ();

	return 0;
}
